----------
--- This work by Evan Leybourn is licensed under a Creative Commons Attribution-ShareAlike 3.0 Australia License 
--- <http://creativecommons.org/licenses/by-sa/3.0/au/>
----------


----------
-- T-SQL (AND STORED PROCEDURE) SYNTAX
--
-- Show that basic T-SQL syntax. While there is a lot more, this is 
-- sufficient to write good test cases
----------
-- Variables
DECLARE @var type;
SET @var = expression;
-- IF/THEN/ELSE
IF cond_1  
	statements_1 (or statement block)
ELSE IF cond_2 
	statements_2 (or statement block)
ELSE 
	statements_n (or statement block)
-- Loops
WHILE condition 
	statement (or statement block)


----------
-- STORED PROCEDURE EXAMPLES
--
-- Withdraw is a simple (and inaccurate) stored procedure updates the 
-- balance in the Accounts table based on the given parameters
----------
--A very simple stored procedure
CREATE PROCEDURE getCustomer(@customerId INT) AS
BEGIN
	SELECT * FROM Customers WHERE customer_id=@customerId
END
-- Call a stored procedure
EXEC getCustomer 1
-- Note: CREATE PROCEDURE Must be the first statement in the block 
-- when executing
DROP PROCEDURE withdraw
CREATE PROCEDURE withdraw @acctNum BIGINT, @amount NUMERIC AS
BEGIN
	DECLARE @current NUMERIC;
	DECLARE @newbalance NUMERIC;

	SELECT @current=balance FROM Accounts 
		WHERE account_id = @acctNum;

	IF @amount > @current
		SELECT 'Insufficient Funds';
	ELSE
		BEGIN
			SET @newbalance = @current - @amount;
			UPDATE Accounts SET balance = @newbalance 
				WHERE account_id = @acctNum;
			SELECT 'New Balance: ' + cast(@newbalance as varchar);
		END
END
-- Call a stored procedure
EXEC withdraw 1, 10;


----------
-- STORED PROCEDURE TEST CASES 
--
-- Exercises on how to wrap a stored procedure in a test case
----------
-- Prepare
CREATE TABLE #result (
	result VARCHAR(100)
)
--Q1. Withdraw SP returns New Balance XXX when withdrawing $1 
--    from User 1
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, 1
SELECT 
	CASE 
		WHEN count(*) = 1
			THEN 'Test Pass'
		ELSE 'Test Fail'
	END
	FROM (
		SELECT * FROM #result WHERE result LIKE 'New Balance%'
	) as t
--Q2. SP returns Insufficient Balance when withdrawing $10000 
--    from User 1
SET NOCOUNT ON
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, 10000
IF (SELECT count(*) FROM #result WHERE result LIKE 'Insufficient Funds')=0
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'
--Q3. SP returns an error when withdrawing -$10
SET NOCOUNT ON
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, -10
IF (SELECT count(*) FROM #result WHERE result LIKE 'Error%')=0
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'
--Q4. SP returns an error when withdrawing from User 100
DELETE FROM #result
INSERT INTO #result EXEC withdraw 100, 1
SELECT 
	CASE 
		WHEN count(*) > 0 THEN 'Test Pass'
		ELSE 'Test Fail: Expected 1 record, found ' + cast(count(*) as varchar)
	END
	FROM (
		SELECT * FROM #result WHERE result LIKE 'Error%'
	) as t
--Q5. SP returns New Balance when withdrawing FLOAT
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, 1.5
DECLARE @msg varchar(255)
DECLARE @result varchar(255)
SET @result = 0
SET @msg = ''
SET @result = (SELECT * FROM #result WHERE result LIKE 'New Balance%')
IF count(@result) = 0
	BEGIN
		SET @msg = 'Test Fail: ' + cast(@result as varchar) + ''
		RAISERROR (@msg , 11, 100)
	END
ELSE
	BEGIN
		SET @msg = 'Test Pass: ' + cast(@result as varchar) + ''
		PRINT @msg
	END
--Q6. SP returns Bal-1.5 when withdrawing 1.499999999
DECLARE @msg varchar(255)
DECLARE @result numeric(18, 2)
DECLARE @expected numeric(18,2)
SET @result = 0
SET @msg = ''
SET @expected = (SELECT cast(balance as numeric(18, 2))-1.5 
					FROM Accounts 
					WHERE account_id=1)
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, 1.4999999999
SET @result = (SELECT cast(SUBSTRING(result, 14, 100) as numeric(18, 2)) 
					FROM #result )
IF @result != @expected
	BEGIN
		SET @msg = 'Test Fail: Expected ' + cast(@expected as varchar) + ', Got ' + cast(@result as varchar) + ''
		RAISERROR (@msg , 11, 100)
	END
ELSE
	BEGIN
		SET @msg = 'Test Pass: ' + cast(@result as varchar) + ''
		PRINT @msg
	END
--Q7. SP returns error when withdrawing ABC
SET NOCOUNT ON
DELETE FROM #result
BEGIN TRANSACTION
	BEGIN TRY
		INSERT INTO #result EXEC withdraw 1, 'ABC'
		RAISERROR ('Test Fail', 11, 1)
		COMMIT TRANSACTION
	END TRY
	BEGIN CATCH
		PRINT 'Test Pass'
		ROLLBACK TRANSACTION
	END CATCH
--Q8. Account table balance=balance-1 when withdrawing $1
DECLARE @old NUMERIC
DECLARE @new NUMERIC 
SET @old = (SELECT balance FROM Accounts where account_id=1)
EXEC withdraw 1, 1
SET @new = (SELECT balance FROM Accounts where account_id=1)
IF @new = @old-1
	PRINT 'Test Pass'
ELSE
	RAISERROR ('Test Fail', 11, 1)
--Q9. Account table balance=sum (transaction table) when ...
--Q10. Account balance=balance when withdrawing $10000
SELECT * INTO #accountstmp FROM accounts
EXEC withdraw 1, 100000
SELECT 
	CASE 
		WHEN count(*) = 0
			THEN 'Test Pass'
		ELSE 'Test Fail'
	END
	FROM (
		SELECT * FROM Accounts
		EXCEPT
		SELECT * FROM #accountstmp
	) as t
DROP TABLE #accountstmp
--or
DECLARE @old NUMERIC
DECLARE @new NUMERIC 
SET @old = (SELECT balance FROM Accounts where account_id=1)
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, 100000
SET @new = (SELECT balance FROM Accounts where account_id=1)
IF @new = @old
	PRINT 'Test Pass'
ELSE
	RAISERROR ('Test Fail', 11, 1)
--Q11. Account table should not change when withdrawing $-10
SELECT * INTO #accountstmp FROM accounts
DELETE FROM #result
INSERT INTO #result EXEC withdraw 1, -10
SELECT 
	CASE 
		WHEN count(*) = 0
			THEN 'Test Pass'
		ELSE 'Test Fail'
	END
	FROM (
		SELECT * FROM Accounts
		EXCEPT
		SELECT * FROM #accountstmp
	) as t
DROP TABLE #accountstmp


----------
-- TRIGGERS
--
-- How to automatically execute a stored procedure when an insert, 
-- update or delete occurs. Note the "inserted" table.
----------
CREATE TRIGGER withdraw ON Transactions AFTER INSERT AS
BEGIN
	DECLARE @current NUMERIC;
	DECLARE @newbalance NUMERIC;
	DECLARE @amount NUMERIC;
	DECLARE @acctNum BIGINT;
	SELECT @current=balance, @amount=i.transaction_amt, @acctNum=i.account_id
		FROM Accounts a, inserted i 
		WHERE a.account_id = i.account_id;
	IF @amount + @current < 0
		BEGIN
			RAISERROR('Insufficient funds', 1, 1)
			ROLLBACK TRANSACTION
		END
	ELSE
		BEGIN
			SET @newbalance = @current + @amount;
			UPDATE Accounts SET balance = @newbalance 
				WHERE account_id = @acctNum;
		END
END
-- Proof
INSERT INTO Transactions VALUES (1, 100, getdate())
SELECT * FROM Accounts WHERE account_id=1
SELECT * FROM Transactions WHERE account_id=1
-- Proof 
INSERT INTO Transactions VALUES (1, -500, getdate())
SELECT * FROM Accounts WHERE account_id=1
SELECT * FROM Transactions WHERE account_id=1
-- Simple ON DELETE TRIGGER to stop a deletion from occuring
CREATE TRIGGER stopDelete ON Transactions AFTER DELETE AS
BEGIN
	RAISERROR('Cannot Delete transactions.', 17, 1)
	ROLLBACK TRANSACTION
END
-- Proof
DELETE FROM Transactions


----------
-- DYNAMIC SQL
--
-- How to automatically run an SQL query that is stored in a string 
-- or table
----------
-- Simple execution
EXEC ('SELECT * FROM Customers WHERE customer_id=1' )
-- Execute through a look - select customers 1-5
DECLARE @num int
SET @num = 5
DECLARE @sql varchar(max)
SET @sql = ''
WHILE @num > 0
	BEGIN
		SET @sql = @sql + 'SELECT * FROM Customers 
			WHERE customer_id=' + cast(@num as varchar)+ ' '
		IF @num <> 1
			SET @sql = @sql + 'UNION '
		SET @num = @num - 1
	END
EXEC (@sql)
-- Build SQL in SQL
SELECT 
    'select ''' + Table_Schema + ''' as SchemaName, 
    ''' + Table_Name + ''' as ObjectName , 
    ''' + case Table_Type when 'base table' then 'table' else lower(Table_Type) end + ''' as ObjectType , 
    count(*) as Rows from ' + Table_Schema +'.'+ Table_Name + ' ' 
FROM information_schema.tables 
ORDER BY Table_Type;
-- Execute a series of SQL from a SQL query result set (via a cursor)
-- This is very advanced.
DECLARE c CURSOR FOR select 
    'select ''' + Table_Schema + ''' as SchemaName, 
    ''' + Table_Name + ''' as ObjectName , 
    ''' + case Table_Type when 'base table' then 'table' else lower(Table_Type) end + ''' as ObjectType , 
    count(*) as Rows from ' + Table_Schema +'.'+ Table_Name + ' ' 
	from information_schema.tables 
DECLARE @bigsql varchar(max)
SET @bigsql = ''
OPEN c
DECLARE @sql varchar(max)
FETCH NEXT FROM c INTO @sql
WHILE @@FETCH_STATUS = 0
	BEGIN
		SET @bigsql = @bigsql + @sql + ' union all '
		FETCH NEXT FROM c INTO @sql
	END
CLOSE c
DEALLOCATE c
EXEC(@bigsql + @sql)
-- Execute sql that is stored in a table.
DECLARE @var varchar(50)
SET @var = (SELECT * FROM sql)
DECLARE c CURSOR FOR SELECT sql FROM sql
OPEN c
DECLARE @sql varchar(max)
FETCH NEXT FROM c INTO @sql
WHILE @@FETCH_STATUS = 0
	BEGIN
		EXEC (@sql)
		FETCH NEXT FROM c INTO @sql
	END
CLOSE c
DEALLOCATE c
-- Dynamic SQL to look for nulls	
declare @fullsql varchar(max),
        @sql varchar(255)
SET @fullsql = ''
declare curse insensitive cursor for
	SELECT 'SELECT '''+TABLE_NAME+''', '''+COLUMN_NAME+''', count('+COLUMN_NAME+') as nulls, count(*)-count('+COLUMN_NAME+') as notnulls FROM '+TABLE_NAME+' 
	'
		FROM INFORMATION_SCHEMA.COLUMNS
open curse
fetch next from curse into @sql
while @@fetch_status = 0
begin
	IF (@fullsql = '') 
		SET @fullsql = @sql
	ELSE
		SET @fullsql = @fullsql + ' UNION ALL '+ @sql
    fetch next from curse into @sql
end -- while fetch
close curse
deallocate curse
exec(@fullsql)


----------
-- CSV EXPORT 
--
-- How to export data to CSV format. Note: this is not supported by 
-- SQLServer and requires some jumping through hoops.
----------
-- Start by setting the correct text format in the options screen
-- Set SQLCMD Mode from the Query menu
:OUT C:\directories\demo.csv
SET NOCOUNT ON; SELECT * FROM Customers 


----------
-- CSV IMPORT
--
-- How to import data from CSV format. Note: this is not supported 
-- by SQLServer and requires some jumping through hoops.
----------
BULK INSERT Staff FROM 'C:\directories\demo.csv' 
	WITH (
		FIRSTROW = 2,
		FIELDTERMINATOR = ',',
		ROWTERMINATOR = '\n',
		TABLOCK
	)
-- If you want to overrule the IDENTITY column
BULK INSERT Staff FROM 'C:\directories\demo.csv' 
	WITH (
		FIRSTROW = 2,
		FIELDTERMINATOR = ',',
		ROWTERMINATOR = '\n',
		TABLOCK,
		KEEPIDENTITY
	)


----------
-- STATISTICS (Optional)
--
-- Look at the statistics engine and rules to update
----------
UPDATE STATISTICS Customers
-- Database Console Command
DBCC SHOW_STATISTICS (Customers, customerpk)


----------
-- PROFILER (Optional)
--
-- Initialisation from SSMS
-- Save to file/table
-- Show events; select errors>user errors and errors>error log
-- Show output > Run all queries > Show output
-- Show Performance Monitor
-- Show Data Collector and Reports
-- Run all queries
----------